"""Exposes the json data files in the /data/names directory."""
import json
import os
import random

female_names = []
male_names = []
nb_names = []

abs_path = os.path.abspath(os.path.dirname(__file__))
names_dir = os.path.join(abs_path, '/data/names/')


def set_name_lists(ethnicity=None):
    """Set three globally available lists of names."""
    if not ethnicity: ethnicity = random.choice(get_ethnicities())
    print("Loading names from: " + ethnicity)
    filename = names_dir + ethnicity + '.json'

    try:
        with open(filename, 'r') as injson:
            data = json.load(injson)
    except:
        return 'Unable to read from file: ' + filename
    else:
        names = [ tuple(name.split(',')) for name in data ]
        random.shuffle(names)

        global female_names
        female_names = [name for name,gender,*desc in names if gender == 'girl']

        global male_names
        male_names = [name for name,gender,*desc in names if gender == 'boy']

        global nb_names
        nb_names = [name for name,gender,*desc in names if gender == 'boygirl']


def get_ethnicities():
    """Retrieve a list of the ethnicities for which name data was scraped.
    Exclude the file extension for human friendliness.
    """
    ethnicities = [ fname.split('.')[0] for fname in os.listdir(names_dir) ]

    return ethnicities


def print_ethnicities():
    """Print a list of the ethnicities for which name data was scraped."""
    print('\nETHNICITIES REPRESENTED: \n')
    print('\n'.join(get_ethnicities()))
    print("\n")



import nltk
import os
import spacy

from collections import Counter
from numpy import dot
from numpy.linalg import norm

nlp = spacy.load('en_core_web_md')


def get_matches(word, tokens, limit, offset=0):
    """Return words from <tokens> that are most closely related to <word>."""
    return closest(tokens, word_vec(word), limit, offset)


def word_vec(word):
    """Return spaCy's vector for <word>."""
    return nlp.vocab[word].vector


def cosine(vec1, vec2):
    """Compare vectors. Borrowed from A. Parish."""
    if norm(vec1) > 0 and norm(vec2) > 0:
        return dot(vec1, vec2) / (norm(vec1) * norm(vec2))
    else:
        return 0.0


def closest(tokens, search_vec, limit, offset=0):
    """Return the <limit> words from <tokens> whose vectors most closely
    resemble the search_vec. Skip the first <offset> results.
    """
    return sorted(tokens,
                  key=lambda x: cosine(search_vec, word_vec(x)),
                  reverse=True)[offset:offset+limit]



"""Defines a class Pantheon."""
import random
from pantheon.gods import *
from math import ceil

class Pantheon:
    def __init__(self, mother_of_creation, father_of_creation):
        if not mother_of_creation.chromosomes == "XX":
            print('Invalid mother of creation. Must have XX chromosomes.')
        if not father_of_creation.chromosomes == "XY":
            print('Invalid father of creation. Must have XY chromosomes.')
        self.gods = {}
        self.add_god(mother_of_creation)
        self.add_god(father_of_creation)


    def add_god(self, god):
        """Add a god to this Pantheon's gods dictionary."""
        self.gods[god.name] = god


    def get_god(self, name_of_god):
        """Retrieve a god from this Pantheon's gods dictionary."""
        try:
            return self.gods[name_of_god]
        except:
            print('No deity named %s in this pantheon.' % name_of_god)


    def spawn(self, generations):
        """Grow this Pantheon by multiplying Gods."""
        egg_donors = [god for god in self.gods.values() if god.chromosomes == 'XX']
        sperm_donors = [god for god in self.gods.values() if god.chromosomes == 'XY']

        for i in range(generations):
            print("\nGENERATION %d\n" % (i+1))
            gen_xx = []
            gen_xy = []

            for egg_donor in egg_donors:
                sperm_donor = random.choice(sperm_donors)
                brood = self.breed(egg_donor, sperm_donor)

                for child in brood:
                    if child.divinity > human:
                        # divine offspring join the Pantheon
                        self.add_god(child)
                    if child.chromosomes == 'XX':
                        gen_xx.append(child)
                    else:
                        gen_xy.append(child)

            # elder gods leave the breeding pool
            egg_donors = [ed for ed in egg_donors if ed.generation > (i-2)]
            sperm_donors = [sd for sd in sperm_donors if sd.generation > (i-3)]

            # mature offspring join the breeding pool
            egg_donors += gen_xx
            sperm_donors += gen_xy


    def breed(self, egg_donor, sperm_donor):
        """Get it on."""
        offspring = []
        num_children = npchoice([1,2], 1, p=[0.8, 0.2])[0] # 20% chance of twins
        for _ in range(num_children):
            child = God(egg_donor, sperm_donor)
            offspring.append(child)
            send_birth_announcement(egg_donor, sperm_donor, child)

        return offspring


def send_birth_announcement(parent_a, parent_b, child):
    """Convenience method for presentations."""
    print("%s - %s" % (child.name, child.epithet))



"""Defines a class God."""
import pantheon.names as names
import pantheon.tokens as tokens
import random
from numpy.random import choice as npchoice
from pantheon.process import *

# Initialize
if not len(tokens.primary_tokens) > 0 : tokens.set_tokens_lists()
if not len(names.female_names) > 0 : names.set_name_lists()

# Divinity constants
god = 3
demi_god = 2
human = 1
divinities = [god, demi_god, human]
p_divinity = {
    # sum : [p_god, p_demi, p_human]
    6: [0.7, 0.3, 0.0], # god+god     = usually god, sometimes demi
    5: [0.5, 0.5, 0.0], # god+demi    = either god or demi
    4: [0.0, 0.8, 0.2], # d+d OR g+h  = usually demi, sometimes human
    3: [0.0, 0.3, 0.7], # demi+human  = usually human, sometimes demi
    2: [0.0, 0.0, 1.0]  # human+human = always human
}

# Gender constants
male = 'M'
female = 'F'
non_binary = 'NB'
genders = [male, female, non_binary]
valid_chromoomes = ['XX','XY']
p_gender = {
     # chromosomes : [p_male, p_female, p_nb]
    'XX': [0.09, 0.85, 0.06], # usually female, sometimes trans, occasionally NB
    'XY': [0.85, 0.09, 0.06]  # usually male, sometimes trans, occasionally NB
}


class God:

    def __init__(self, egg_donor, sperm_donor, chromosomes=None):
        self.set_chromosomes(chromosomes)
        self.set_gender()
        self.set_inherited_traits(egg_donor, sperm_donor)
        self.set_name()
        self.set_epithet()


    def set_chromosomes(self, chromosomes=None):
        """This model uses the XY sex-determination system. Sex != gender.
        Assign either XX or XY randomly with a 50/50 chance of each, unless
        <chromosomes> are passed as an argument.
        """
        if chromosomes and chromosomes in valid_chromosomes:
            self.chromosomes = chromosomes
        else:
            self.chromosomes = random.choice(['XX','XY'])


    def set_gender(self):
        """This model recognizes that sex chromosomes don't always line up with
        gender. Assign M, F, or NB according to the probabilities in p_gender.
        """
        if not self.chromosomes: self.set_chromosomes()
        self.gender = npchoice(genders, 1, p=p_gender[self.chromosomes])[0]


    def set_inherited_traits(self, egg_donor, sperm_donor):
        """Accept either strings or Gods as inputs."""
        if type(egg_donor) == str:
            self.reproduce_asexually(egg_donor, sperm_donor)
        else:
            self.reproduce_sexually(egg_donor, sperm_donor)


    def reproduce_asexually(self, egg_word, sperm_word):
        """Produce two gametes, an egg and a sperm, from the input strings.
        Combine them to produce a genome a la sexual reproduction.
        """
        egg = self.generate_gamete(egg_word)
        sperm = self.generate_gamete(sperm_word)

        self.genome = list(set(egg + sperm)) # Eliminate duplicates
        self.generation = 1
        self.divinity = god # If it springs out of the ether, it's a god


    def reproduce_sexually(self, egg_donor, sperm_donor):
        """Produce two gametes, an egg and a sperm, from input Gods. Combine
        them to produce a genome a la sexual reproduction. Assign divinity
        according to probabilities in p_divinity. The more divine the parents,
        the more divine their offspring.
        """
        egg_word = random.choice(egg_donor.genome)
        egg = self.generate_gamete(egg_word)
        sperm_word = random.choice(sperm_donor.genome)
        sperm = self.generate_gamete(sperm_word)

        self.genome = list(set(egg + sperm)) # Eliminate duplicates
        self.parents = [egg_donor, sperm_donor]
        self.generation = max(egg_donor.generation, sperm_donor.generation) + 1
        sum_ = egg_donor.divinity + sperm_donor.divinity
        self.divinity = npchoice(divinities, 1, p=p_divinity[sum_])[0]


    def set_name(self):
        """Pick a random name from the lists loaded with the model. For Gods that
        identify as neither M nor F, the model attempts to retrieve an androgynous
        name. Note: not all of the scraped name lists contain androgynous names.
        """
        if not self.gender: self.set_gender()

        name = ''
        if self.gender == female:
            name = names.female_names.pop()
        elif self.gender == male:
            name = names.male_names.pop()
        else:
            try:
                name = names.nb_names.pop()
            except:
                # No androgynous names available
                name = male_names.pop()

        self.name = name


    def set_epithet(self):
        """Divine an appropriate epithet for this God. (See what I did there?)"""
        if self.divinity == human:
            obsession = random.choice(self.genome)
            if self.gender == female:
                self.epithet = 'an ordinary woman'
            elif self.gender == male:
                self.epithet = 'an ordinary man'
            else:
                self.epithet = 'an ordinary human being who loves ' + obsession
            return # Return early. The rest of the function deals with gods.

        if self.gender == female:
            title = 'Goddess'
        elif self.gender == male:
            title = 'God'
        else:
            title = 'Divine Being'
        if self.divinity == demi_god:
            title = 'Semi-' + title if self.gender == non_binary else 'Demi-' + title

        num_domains = npchoice([1,2,3,4], 1, p=[0.05, 0.35, 0.55, 0.05])[0]

        if num_domains == 1:
            template = '%s of %s'
        if num_domains == 2:
            template = '%s of %s and %s'
        elif num_domains == 3:
            template = '%s of %s, %s, and %s' # Oxford comma, the most divine punctuation.
        elif num_domains == 4:
            template = '%s of %s, %s, %s, and %s'

        self.domains = random.sample(self.genome, num_domains)

        # Put it all together
        self.epithet = template % (title, *self.domains)


    def generate_gamete(self, egg_or_sperm_word):
        """Extract 23 'chromosomes' aka words from 'gene pool' aka list of tokens
        by searching the list of tokens for words that are related to the given
        egg_or_sperm_word.
        """
        p_rate_of_mutation = [0.9, 0.1]
        should_use_mutant_pool = (npchoice([0,1], 1, p=p_rate_of_mutation)[0] == 1)
        if should_use_mutant_pool:
            pool = tokens.secondary_tokens
        else:
            pool = tokens.primary_tokens

        return get_matches(egg_or_sperm_word, pool, 23)


    def print_parents(self):
        """Print parents' names and epithets."""
        if self.gender == female:
            title = 'Daughter'
        elif self.gender == male:
            title = 'Son'
        else:
            title = 'Child'

        p1 = self.parents[0]
        p2 = self.parents[1]

        template = '%s of %s, the %s, and %s, the %s.'

        print(template % (title, p1.name, p1.epithet, p2.name, p2.epithet))



"""Exposes the json data files in the /data/tokens directory. These files contain
raw tokens lists pulled from texts in the /data/corpora directory.
"""
import json
import nltk
import os
import random

primary_tokens = [] # tokens that represent "normal" genes
secondary_tokens = [] # tokens that represent "mutant" genes


abs_path = os.path.abspath(os.path.dirname(__file__))
corpora_dir = os.path.join(abs_path, '/data/corpora/')
tokens_dir = os.path.join(abs_path, '/data/tokens/')

def sequence_genes():
    """An alias."""
    tokenize_corpora()


def filter_genes(pool, filters):
    """An alias."""
    save_tokens_to_dir(pool, filters)


def define_gene_pool(pool, individuals):
    """An alias."""
    make_tokens_dir(pool, individuals)


def select_gene_pool(dir_):
    """An alias."""
    set_token_lists(dir_)


def tokenize_texts():
    """Generate a json file for each txt file in the /data/corpora directory."""
    text_files = [fname for fname in os.listdir(corpora_dir) \
        if fname.split('.')[1] == 'txt']

    for text_fname in text_files:
        json_fname = text_fname.split('.')[0] + '.json'
        if os.path.isfile(corpora_dir + json_fname):
            continue # already tokenized

        print("Tokenizing " + text_fname)
        text = open(corpora_dir + text_fname).read()
        words = nltk.word_tokenize(text)
        with open(corpora_dir + json_fname, 'w') as outjson:
            json.dump(words, outjson)


def list_tokenized_texts():
    """Retrieve the filenames of all tokenized text files in /data/corpora.
    Useful when you want to hand-pick <sources> for make_tokens_dir().
    """
    return [f for f in os.listdir(corpora_dir) if f.split('.')[1] == 'json']


def make_tokens_dir(dir_, sources):
    """Create a new directory named <dir_>. Create a new file within it called
    sources.json. The input <sources> is a list of names of tokenized texts.
    Write <sources> into sources.json.
    """
    os.mkdir(tokens_dir + dir_)
    for source in sources:
        if not os.path.isfile(corpora_dir + source):
            print('Invalid source: ' + source)
            return

    with open(tokens_dir + dir_ + '/sources.json', 'w') as outjson:
        json.dump(sources, outjson)


def make_tokens_list(dir_, filters):
    """Find sources.json in <dir_>. It contains a list of tokenized texts. For
    each tokenized text listed in sources.json, read its tokens, filter them,
    and add them to an aggregated list. Write the aggregated list to disk using
    a filename based on the <filters> given.
    """
    with open(tokens_dir + dir_ + '/sources.json', 'r') as injson:
        data = json.load(injson)
        sources = [corpora_dir + fname for fname in data]

    with open('data/skipwords.txt', 'r') as f:
        skipwords = [line.rstrip() for line in f]

    tokens_list = []
    for fname in sources:
        print("Incorporating tokens from " + fname)
        with open(fname, 'r') as injson:
            data = json.load(injson)
            words = [w.lower() for w in data if not w == '']
            filtered = [w for w,p in nltk.pos_tag(words) if p in filters]
            sanitized = [w for w in filtered if not w in skipwords]
            tokens_list += sanitized

    tokens_list = list(set(tokens_list)) # unique
    target = tokens_dir + dir_ + '/' + '-'.join(filters) + '.json'
    with open(target, 'w') as outjson:
        json.dump(tokens_list, outjson)


def set_tokens_lists(dir_=None):
    if not dir_: dir_ = random.choice(os.listdir(tokens_dir))
    print("Loading tokens from: " + dir_)

    p_fname = tokens_dir + dir_ + '/NNS.json'
    s_fname = tokens_dir + dir_ + '/VBG.json'

    global primary_tokens
    primary_tokens = json.load(open(p_fname, 'r'))

    global secondary_tokens
    secondary_tokens = json.load(open(s_fname, 'r'))



"""A setuptools based setup module."""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    # This is the name of your project.
    # It will live at https://pypi.org/project/pantheon-generator/
    # $ pip install pantheon-generator
    name='pantheon-generator',  # Required

    version='0.9.2',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='A module for generating lists of related dieties',  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # Denotes that our long_description is in Markdown; valid values are
    # text/plain, text/x-rst, and text/markdown
    #
    # This field corresponds to the "Description-Content-Type" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type='text/x-rst',  # Optional (see note above)

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/carawarner/pantheon',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Cara Warner',  # Optional

    license='Apache License 2.0',

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='procgenitor@gmail.com',  # Optional

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    keywords='procgen nlg creativity',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['nltk', 'numpy', 'spacy'],  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    # extras_require={  # Optional
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    package_data={  # Optional
        'pantheon': ['data/*.txt', 'data/names/*.json', 'data/tokens/*/*.json'],
    },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    # entry_points={  # Optional
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/carawarner/pantheon/issues',
        'Source': 'https://github.com/carawarner/pantheon',
    },
)
